@@ -13,6 +13,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import logging
+from oslo_log import fixture as log_fixture
+
 import cProfile
 from eventlet import queue
 import mock
@@ -21,15 +24,16 @@ from pstats import Stats
 from neutron_lib.api.definitions import portbindings
 from neutron_lib import constants as n_const
 import neutron_lib.context
+from neutron_lib.db import api as db_api
 from neutron_lib.plugins import constants as p_const
 from neutron_lib.plugins import directory
+from neutron_lib.services.trunk import constants as trunk_const
 from oslo_config import cfg
 
 from neutron.api.rpc.handlers import l3_rpc
-from neutron.common import constants as c_const
 from neutron.common import utils as common_utils
 from neutron.plugins.ml2.drivers import type_vxlan  # noqa
-from neutron.services.trunk import constants as trunk_const
+from neutron.plugins.ml2 import models as ml2_models
 from neutron.tests.common import helpers
 from neutron.tests.unit.plugins.ml2 import test_plugin
 
@@ -39,6 +43,7 @@ from networking_arista.tests.unit import utils
 
 
 ENABLE_PROFILER = False
+ENABLE_LOGGING = False
 
 
 class MechTestBase(test_plugin.Ml2PluginV2TestCase):
@@ -57,6 +62,10 @@ class MechTestBase(test_plugin.Ml2PluginV2TestCase):
         return p
 
     def setUp(self):
+        if ENABLE_LOGGING:
+            log_fixture.setlevel.logging.root.setLevel(logging.DEBUG)
+            self.useFixture(log_fixture.SetLogLevel(['neutron'],
+                                                    logging.DEBUG))
         if ENABLE_PROFILER:
             self.pr = cProfile.Profile()
             self.pr.enable()
@@ -268,16 +277,36 @@ class MechTestBase(test_plugin.Ml2PluginV2TestCase):
                    portbindings.VIF_TYPE: portbindings.VIF_TYPE_UNBOUND}}
         self.plugin.update_port(self.context, port_id, p_dict)
 
-    def bind_dvr_to_host(self, port, host):
-        p_dict = {'port':
-                  {'device_id': port['device_id'],
-                   'device_owner': port['device_owner'],
-                   portbindings.HOST_ID: host}}
-        self.plugin.update_distributed_port_binding(self.context,
-                                                    port['id'], p_dict)
-        self.plugin.update_port_status(self.context, port['id'],
-                                       n_const.PORT_STATUS_ACTIVE,
-                                       host)
+    def bind_dvr_to_host(self, port, host, notify_ml2=True, seg_id=None):
+        if notify_ml2:
+            p_dict = {'port':
+                      {'device_id': port['device_id'],
+                       'device_owner': port['device_owner'],
+                       portbindings.HOST_ID: host}}
+            self.plugin.update_distributed_port_binding(self.context,
+                                                        port['id'], p_dict)
+            self.plugin.update_port_status(self.context, port['id'],
+                                           n_const.PORT_STATUS_ACTIVE,
+                                           host)
+        else:
+            session = db_api.get_writer_session()
+            with session.begin():
+                distributed_binding = {'port_id': port['id'],
+                                       'host': host,
+                                       'router_id': port['device_id'],
+                                       'status': 'ACTIVE',
+                                       'vif_type': 'distributed',
+                                       'vnic_type': portbindings.VNIC_NORMAL,
+                                       'profile': '',
+                                       'vif_details': ''}
+                session.add(ml2_models.DistributedPortBinding(
+                    **distributed_binding))
+                binding_level = {'port_id': port['id'],
+                                 'host': host,
+                                 'segment_id': seg_id,
+                                 'level': 0,
+                                 'driver': 'ovs'}
+                session.add(ml2_models.PortBindingLevel(**binding_level))
         p_ctx = self.plugin.get_bound_port_context(self.context, port['id'],
                                                    host)
         return port, p_ctx
@@ -286,6 +315,10 @@ class MechTestBase(test_plugin.Ml2PluginV2TestCase):
         self.plugin.update_port_status(
             self.context, port['id'], n_const.PORT_STATUS_DOWN, host)
 
+    def set_port_to_error_state(self, port):
+        self.plugin.update_port_status(
+            self.context, port['id'], n_const.PORT_STATUS_ERROR)
+
     def delete_port(self, port_id):
         self.plugin.delete_port(self.context, port_id)
 
@@ -451,7 +484,7 @@ class L3HARouterTestFramework(MechTestBase):
 
     def get_ha_network(self, router):
         networks = [getattr(net, 'Network') for net in db_lib.get_networks()]
-        HA_network_name = c_const.HA_NETWORK_NAME % router['project_id']
+        HA_network_name = n_const.HA_NETWORK_NAME % router['project_id']
         net_id = [net.id for net in networks if net.name == HA_network_name]
         return net_id[0]
 
@@ -477,7 +510,7 @@ class L3HARouterTestFramework(MechTestBase):
 
     def assertL3HANetworkCreated(self, router, net_id):
         endpoint = self._get_endpoint('network')
-        HA_network_name = c_const.HA_NETWORK_NAME % router['project_id']
+        HA_network_name = n_const.HA_NETWORK_NAME % router['project_id']
 
         def resource_created():
             cvx_data = self.cvx.endpoint_data[endpoint]
@@ -509,7 +542,7 @@ class L3HARouterTestFramework(MechTestBase):
 
     def assertL3HAPortCreated(self, router, port_id):
         endpoint = self._get_endpoint('router_port')
-        port_name = c_const.HA_PORT_NAME % router['project_id']
+        port_name = n_const.HA_PORT_NAME % router['project_id']
 
         def resource_created():
             cvx_data = self.cvx.endpoint_data[endpoint]
